from overhave.entities import StepPrefixesModel

RUSSIAN_PREFIXES = StepPrefixesModel(
    FEATURE="Функция:",
    SCENARIO_OUTLINE="Структура сценария:",
    SCENARIO="Сценарий:",
    BACKGROUND="Предыстория:",
    EXAMPLES="Примеры:",
    EXAMPLES_VERTICAL="Примеры: Вертикально",
    GIVEN="Дано ",
    WHEN="Когда ",
    THEN="То ",
    AND="И ",
    BUT="Но ",
)



from datetime import timedelta
from pathlib import Path

import click

from overhave.base_settings import LoggingSettings
from overhave.cli.group import overhave
from overhave.transport import OverhaveS3Bucket, OverhaveS3ManagerSettings, S3Manager
from overhave.utils import get_current_time


@overhave.group(short_help="Run s3 cloud interaction commands")
def s3() -> None:
    pass


@s3.group(short_help="S3 cloud bucket's interaction commands")
def bucket() -> None:
    pass


def _check_bucket_registered(name: str) -> None:
    if name in (item.value for item in list(OverhaveS3Bucket)):
        return
    click.secho(f"Note: specified s3 bucket name '{name}' not presented in OverhaveS3Bucket enum!", fg="yellow")


def _get_s3_manager() -> S3Manager:
    LoggingSettings().setup_logging()
    manager = S3Manager(OverhaveS3ManagerSettings(autocreate_buckets=False))
    manager.initialize()
    return manager


@bucket.command(short_help="Create s3 cloud bucket")
@click.option(
    "-n",
    "--name",
    type=str,
    help="Declared s3 bucket",
)
def create(name: str) -> None:
    """Create s3 bucket."""
    _check_bucket_registered(name)
    _get_s3_manager().create_bucket(name)


@bucket.command(short_help="Delete s3 cloud bucket")
@click.option(
    "-n",
    "--name",
    type=str,
    help="Declared s3 bucket",
)
@click.option(
    "-f",
    "--force",
    is_flag=True,
    help="Delete all files in bucket, then delete bucket",
)
def delete(name: str, force: bool) -> None:
    """Delete s3 bucket."""
    _check_bucket_registered(name)
    _get_s3_manager().delete_bucket(name, force=force)


@bucket.command(short_help="Remove old s3 cloud bucket files")
@click.option(
    "-n",
    "--name",
    type=str,
    help="Declared s3 bucket",
)
@click.option(
    "-d",
    "--days",
    type=int,
    help="Remove all files in bucket older then specified days value",
)
def remove_files(name: str, days: int) -> None:
    """Remove s3 bucket files older than specified number of days."""
    _check_bucket_registered(name)
    manager = _get_s3_manager()
    target_date = get_current_time() - timedelta(days=days)

    objects = manager.get_bucket_objects(name)
    objects_to_delete = []
    for obj in objects:
        if not obj.modified_at < target_date:
            continue
        objects_to_delete.append(obj)
    if not objects_to_delete:
        click.secho(f"No one object older than {days} days.")
        return
    click.secho(f"Objects older then {days} days: {[x.name for x in objects_to_delete]}")
    manager.delete_bucket_objects(bucket=name, objects=objects_to_delete)


@s3.command(short_help="Download file from s3 bucket")
@click.option(
    "-b",
    "--bucket",
    type=str,
    help="Declared s3 bucket",
)
@click.option(
    "-f",
    "--filename",
    type=str,
    help="Filename for downloading",
)
@click.option("-d", "--dir-to-save", type=str, help="Directory for saving file", default=".")
def download_file(bucket: str, filename: str, dir_to_save: str) -> None:
    """Create s3 bucket."""
    _check_bucket_registered(bucket)
    _get_s3_manager().download_file(filename=filename, bucket=bucket, dir_to_save=Path(dir_to_save))



# flake8: noqa
from .base_factory import IOverhaveFactory
from .components import (
    AdminFactory,
    EmulationFactory,
    IAdminFactory,
    IEmulationFactory,
    IPublicationFactory,
    ISynchronizerFactory,
    ITaskConsumerFactory,
    ITestExecutionFactory,
    PublicationFactory,
    TestExecutionFactory,
)
from .consumer_factory import ConsumerFactory
from .context import (
    OverhaveAdminContext,
    OverhaveEmulationContext,
    OverhavePublicationContext,
    OverhaveSynchronizerContext,
    OverhaveTestExecutionContext,
    TApplicationContext,
)
from .getters import (
    get_admin_factory,
    get_emulation_factory,
    get_publication_factory,
    get_synchronizer_factory,
    get_test_execution_factory,
)



# flake8: noqa
from .admin import admin
from .consumers import consumer
from .db import db, set_config_to_context
from .group import overhave
from .s3 import s3
from .synchronization import synchronize



# flake8: noqa
from .formatters import (
    datetime_formatter,
    draft_feature_formatter,
    draft_prurl_formatter,
    draft_testrun_formatter,
    feature_link_formatter,
    json_formatter,
    result_report_formatter,
    task_formatter,
)



import click


def _git_sync() -> None:
    pass


@click.command(short_help="Synchronize features with actual state of git repository on default branch")
@click.pass_obj
def git_sync() -> None:
    """Create all metadata tables."""
    _git_sync()



import logging
from json import JSONDecodeError
from typing import Any, Dict, Generic, Mapping, Optional, Union, cast

import requests
import tenacity
from pydantic import BaseModel, ValidationError
from pydantic.main import ModelMetaclass
from requests.auth import AuthBase
from yarl import URL

from overhave.transport.http.base_client.objects import HttpMethod
from overhave.transport.http.base_client.settings import HttpSettingsType

logger = logging.getLogger(__name__)


class BaseHttpClientException(Exception):
    """Base exception for BaseHttpClient."""


class HttpClientValidationError(BaseHttpClientException):
    """Model validation error for BaseHttpClient."""


class BaseHttpClient(Generic[HttpSettingsType]):
    """Base client for HTTP communications."""

    def __init__(self, settings: HttpSettingsType) -> None:
        self._settings = settings

    @staticmethod
    def _parse_or_raise(response: requests.Response, model: ModelMetaclass) -> BaseModel:
        try:
            return cast(BaseModel, model).parse_obj(response.json())
        except (ValueError, ValidationError, JSONDecodeError) as e:
            url = getattr(response, "raw_url", response.url)
            raise HttpClientValidationError(f'Response validation error for "{url}"') from e

    @tenacity.retry(
        reraise=True,
        retry=tenacity.retry_if_exception_type(requests.ConnectionError),
        stop=tenacity.stop_after_attempt(3),
        before_sleep=tenacity.before_sleep_log(logger, logger.level),
        after=tenacity.after_log(logger, logger.level),
    )
    def _make_request(
        self,
        method: HttpMethod,
        url: URL,
        params: Optional[Dict[str, Any]] = None,
        json: Optional[Dict[str, Any]] = None,
        data: Optional[Union[str, bytes, Mapping[Any, Any]]] = None,
        auth: Optional[AuthBase] = None,
        raise_for_status: bool = True,
    ) -> requests.Response:
        response = requests.request(
            method=method.value,
            url=url.human_repr(),
            params=params,
            json=json,
            data=data,
            auth=auth,
            timeout=self._settings.timeout,
        )
        if raise_for_status:
            response.raise_for_status()
        return response



from functools import cache

from _pytest.nodes import Item

from overhave.pytest_plugin.helpers.allure_utils.description_manager import DescriptionManager
from overhave.pytest_plugin.helpers.allure_utils.step_context_runner import StepContextRunner
from overhave.pytest_plugin.helpers.basic import get_scenario


def add_scenario_title_to_report(item: Item) -> None:
    item._obj.__allure_display_name__ = get_scenario(item).name  # type: ignore


@cache
def get_step_context_runner() -> StepContextRunner:
    from overhave.factory import get_test_execution_factory

    return StepContextRunner(settings=get_test_execution_factory().context.step_context_settings)


@cache
def get_description_manager() -> DescriptionManager:
    from overhave.factory import get_test_execution_factory

    return DescriptionManager(settings=get_test_execution_factory().context.description_manager_settings)



import logging
from typing import Optional

from flask import redirect
from flask_login import LoginManager
from werkzeug import Response

from overhave.entities import SystemUserModel
from overhave.storage import ISystemUserStorage

logger = logging.getLogger(__name__)


class FlaskLoginManager(LoginManager):
    """Custom Flask LoginManager for user acknowledgement."""

    def __init__(self, system_user_storage: ISystemUserStorage, login_view: str) -> None:
        super().__init__()
        self.login_view = login_view
        self.user_loader(self._get_user)
        self.unauthorized_handler(self._unathorized_response)

        self._system_user_storage = system_user_storage

    def _get_user(self, user_id: int) -> Optional[SystemUserModel]:
        logger.info("Get user by id=%s...", user_id)
        return self._system_user_storage.get_user(user_id=user_id)

    def _unathorized_response(self) -> Response:
        return redirect(f"/{self.login_view}")



from typing import Any, Dict, Optional

from pydantic import validator
from yarl import URL

from overhave.transport.http import BaseHttpClientSettings


class TokenizerClientSettings(BaseHttpClientSettings):
    """Important environments and settings for :class:`TokenizerClient`."""

    enabled: bool = False
    url: Optional[URL] = None  # type: ignore
    initiator: str = "Overhave"
    remote_key: Optional[str] = None
    remote_key_name: Optional[str] = None

    class Config:
        env_prefix = "OVERHAVE_GITLAB_TOKENIZER_"

    @validator("url", "remote_key", "remote_key_name")
    def validate_remote_key_and_initiator(cls, v: Optional[str], values: Dict[str, Any]) -> Optional[str]:
        if values.get("enabled") and not isinstance(v, str):
            raise ValueError("Please verify that url, remote_key and remote_key_name variables are not nullable!")
        return v



import abc
from functools import cached_property

from overhave.factory.base_factory import IOverhaveFactory
from overhave.factory.components.abstract_consumer import ITaskConsumerFactory
from overhave.factory.components.s3_init_factory import FactoryWithS3ManagerInit
from overhave.factory.context import OverhaveTestExecutionContext
from overhave.test_execution.executor import ITestExecutor, TestExecutor
from overhave.transport import TestRunTask


class ITestExecutionFactory(IOverhaveFactory[OverhaveTestExecutionContext], ITaskConsumerFactory[TestRunTask], abc.ABC):
    """Abstract factory for Overhave test execution application."""

    @property
    @abc.abstractmethod
    def test_executor(self) -> ITestExecutor:
        pass


class TestExecutionFactory(FactoryWithS3ManagerInit[OverhaveTestExecutionContext], ITestExecutionFactory):
    """Factory for Overhave test execution application."""

    context_cls = OverhaveTestExecutionContext

    @cached_property
    def _test_executor(self) -> ITestExecutor:
        return TestExecutor(
            file_settings=self.context.file_settings,
            feature_storage=self._feature_storage,
            scenario_storage=self._scenario_storage,
            test_run_storage=self._test_run_storage,
            file_manager=self._file_manager,
            test_runner=self._test_runner,
            report_manager=self._report_manager,
        )

    @property
    def test_executor(self) -> ITestExecutor:
        return self._test_executor

    def process_task(self, task: TestRunTask) -> None:
        return self._test_executor.process_test_task(task)



import abc
from typing import List

from overhave import db


class ISystemUserGroupStorage(abc.ABC):
    """Abstract class for system user storage."""

    @staticmethod
    @abc.abstractmethod
    def has_any_group(user_groups: List[str]) -> bool:
        pass


class SystemUserGroupStorage(ISystemUserGroupStorage):
    """Class for system user storage."""

    @staticmethod
    def has_any_group(user_groups: List[str]) -> bool:
        with db.create_session() as session:
            group = session.query(db.GroupRole).filter(db.GroupRole.group.in_(user_groups)).first()
            return group is not None



# flake8: noqa
from .draft_storage import DraftStorage, IDraftStorage, UniqueDraftCreationError
from .emulation_storage import EmulationStorage, EmulationStorageError, IEmulationStorage
from .feature_storage import FeatureStorage, IFeatureStorage
from .feature_tag_storage import FeatureTagStorage, IFeatureTagStorage
from .feature_type_storage import FeatureTypeStorage, IFeatureTypeStorage
from .scenario_storage import IScenarioStorage, ScenarioStorage
from .system_user_group_storage import ISystemUserGroupStorage, SystemUserGroupStorage
from .system_user_storage import ISystemUserStorage, SystemUserStorage
from .test_run_storage import ITestRunStorage, TestRunStorage



import logging
import socket
from logging.config import DictConfigurator  # type: ignore
from typing import Any, Dict, Optional

from pydantic import BaseSettings, Field, validator
from sqlalchemy import engine_from_config
from sqlalchemy.engine import Engine
from sqlalchemy.engine.url import URL, make_url
from sqlalchemy.exc import ArgumentError
from sqlalchemy.pool import SingletonThreadPool


class BaseOverhavePrefix(BaseSettings):
    """Possibility to change Overhave default settings from environment."""

    class Config:
        env_prefix = "OVERHAVE_"


class SAUrl(URL):
    """Custom SQLAlchemy URL for Pydantic BaseSettings validation."""

    @classmethod
    def __get_validators__(cls):  # type: ignore
        yield cls.validate

    @classmethod
    def validate(cls, v: str) -> URL:
        try:
            return make_url(v)
        except ArgumentError as e:
            raise ValueError from e


class DataBaseSettings(BaseOverhavePrefix):
    """Overhave database settings."""

    db_url: SAUrl = Field(SAUrl.validate("postgresql://postgres:postgres@localhost/overhave"))
    db_pool_recycle: int = 500
    db_pool_size: int = 6
    db_echo: bool = False
    db_application_name: str = socket.gethostname()
    db_connect_timeout: int = 30

    def create_engine(self) -> Engine:
        return engine_from_config(
            {
                "url": self.db_url,
                "pool_recycle": self.db_pool_recycle,
                "pool_pre_ping": True,
                "pool_size": self.db_pool_size,
                "poolclass": SingletonThreadPool,
                "connect_args": {
                    "connect_timeout": self.db_connect_timeout,
                    "application_name": self.db_application_name,
                },
            },
            prefix="",
        )

    def setup_db(self) -> None:
        from overhave.db.base import metadata

        metadata.bind = self.create_engine()


class LoggingSettings(BaseOverhavePrefix):
    """Overhave logging settings."""

    log_level: str = logging.getLevelName(logging.INFO)
    log_config: Dict[str, Any] = {}

    @validator("log_config", each_item=True)
    def dict_config_validator(cls, v: Dict[str, Any]) -> Optional[DictConfigurator]:
        if not v:
            return None
        return DictConfigurator(v)

    def setup_logging(self) -> None:
        if isinstance(self.log_config, DictConfigurator):
            self.log_config.configure()
        else:
            logging.basicConfig(level=self.log_level)



import gitlab

from overhave.transport.http.gitlab_client.objects import TokenType


class InvalidTokenTypeError(Exception):
    """Error for choosing invalid token type value."""


def get_gitlab_python_client(url: str, token_type: TokenType, token: str) -> gitlab.Gitlab:
    if token_type is TokenType.OAUTH:
        return gitlab.Gitlab(url, oauth_token=token)
    if token_type is TokenType.PRIVATE:
        return gitlab.Gitlab(url, private_token=token)
    if token_type is TokenType.JOB:
        return gitlab.Gitlab(url, private_token=token)
    raise InvalidTokenTypeError(
        f"Please, verify your token type! {token_type} not in {[TokenType.OAUTH, TokenType.JOB, TokenType.PRIVATE]}"
    )
